searchNodes=[{"ref":"Raxx.html","title":"Raxx","type":"module","doc":"Tooling to work with HTTP. Several data structures are defined to model parts of the communication between client and server. Raxx.Request: metadata sent by a client before sending content. Raxx.Response: metadata sent by a server before sending content. Raxx.Data: A part of a messages content. Raxx.Tail: metadata set by client or server to conclude communication. This module contains functions to create and manipulate these structures. See Raxx.Server for implementing a web application."},{"ref":"Raxx.html#complete?/1","title":"Raxx.complete?/1","type":"function","doc":"Does the message struct contain all the data to be sent. Examples iex&gt; request(:GET, &quot;/&quot;) ...&gt; |&gt; complete?() true iex&gt; response(:ok) ...&gt; |&gt; set_body(&quot;Hello, World!&quot;) ...&gt; |&gt; complete?() true iex&gt; response(:ok) ...&gt; |&gt; set_body(true) ...&gt; |&gt; complete?() false"},{"ref":"Raxx.html#data/1","title":"Raxx.data/1","type":"function","doc":"Construct a Raxx.Data Wrap a piece of data as being part of a message&#39;s body. Examples iex&gt; data(&quot;Hi&quot;).data &quot;Hi&quot;"},{"ref":"Raxx.html#delete_header/2","title":"Raxx.delete_header/2","type":"function","doc":"Delete a header, if present from a request or response."},{"ref":"Raxx.html#fetch_query/1","title":"Raxx.fetch_query/1","type":"function","doc":"This function never returns an error use, get_query/1 instead."},{"ref":"Raxx.html#get_content_length/1","title":"Raxx.get_content_length/1","type":"function","doc":"Get the integer value for the content length of a message. A well formed struct will always have a non negative content length, or none. Examples iex&gt; response(:ok) ...&gt; |&gt; set_content_length(0) ...&gt; |&gt; get_content_length() 0 iex&gt; response(:ok) ...&gt; |&gt; get_content_length() nil"},{"ref":"Raxx.html#get_header/3","title":"Raxx.get_header/3","type":"function","doc":"Get the value of a header field. Examples iex&gt; response(:ok) ...&gt; |&gt; set_header(&quot;content-type&quot;, &quot;text/html&quot;) ...&gt; |&gt; get_header(&quot;content-type&quot;) &quot;text/html&quot; iex&gt; response(:ok) ...&gt; |&gt; set_header(&quot;content-type&quot;, &quot;text/html&quot;) ...&gt; |&gt; get_header(&quot;location&quot;) nil iex&gt; response(:ok) ...&gt; |&gt; set_header(&quot;content-type&quot;, &quot;text/html&quot;) ...&gt; |&gt; get_header(&quot;content-type&quot;, &quot;text/plain&quot;) &quot;text/html&quot; iex&gt; response(:ok) ...&gt; |&gt; set_header(&quot;content-type&quot;, &quot;text/html&quot;) ...&gt; |&gt; get_header(&quot;location&quot;, &quot;/&quot;) &quot;/&quot;"},{"ref":"Raxx.html#get_query/1","title":"Raxx.get_query/1","type":"function","doc":"Return the decoded query from a request A map is always returned, even in the case of a request without a query string. Examples iex&gt; request(:GET, &quot;/&quot;) ...&gt; |&gt; get_query() %{} iex&gt; request(:GET, &quot;/?&quot;) ...&gt; |&gt; get_query() %{} iex&gt; request(:GET, &quot;/?foo=bar&quot;) ...&gt; |&gt; get_query() %{&quot;foo&quot; =&gt; &quot;bar&quot;}"},{"ref":"Raxx.html#idempotent?/1","title":"Raxx.idempotent?/1","type":"function","doc":"Effect of handling this request more than once should be identical to handling it once. For full definition of idempotent methods see RFC7231 Examples iex&gt; request(:GET, &quot;/&quot;) ...&gt; |&gt; idempotent?() true iex&gt; request(:POST, &quot;/&quot;) ...&gt; |&gt; idempotent?() false iex&gt; request(:PUT, &quot;/&quot;) ...&gt; |&gt; idempotent?() true"},{"ref":"Raxx.html#normalized_path/1","title":"Raxx.normalized_path/1","type":"function","doc":""},{"ref":"Raxx.html#reason_phrase/1","title":"Raxx.reason_phrase/1","type":"function","doc":"The RFC7231 specified reason phrase for each known HTTP status code. Examples iex&gt; reason_phrase(200) &quot;OK&quot; iex&gt; reason_phrase(500) &quot;Internal Server Error&quot; iex&gt; reason_phrase(999) nil"},{"ref":"Raxx.html#redirect/2","title":"Raxx.redirect/2","type":"function","doc":"Create a response to redirect client to the given url. Response status can be set using the :status option. The body can be set with the :body option. Content-type header can be set with the :content_type option. Examples iex&gt; redirect(&quot;/foo&quot;) ...&gt; |&gt; get_header(&quot;location&quot;) &quot;/foo&quot; iex&gt; redirect(&quot;/foo&quot;) ...&gt; |&gt; Map.get(:body) false iex&gt; redirect(&quot;/foo&quot;) ...&gt; |&gt; Map.get(:status) 303 iex&gt; redirect(&quot;/foo&quot;, body: &quot;Redirecting...&quot;) ...&gt; |&gt; Map.get(:body) &quot;Redirecting...&quot; iex&gt; redirect(&quot;/foo&quot;, body: &quot;Redirecting...&quot;) ...&gt; |&gt; get_header(&quot;content-type&quot;) nil iex&gt; redirect(&quot;/foo&quot;, body: &quot;Redirecting...&quot;, content_type: &quot;text/html&quot;) ...&gt; |&gt; get_header(&quot;content-type&quot;) &quot;text/html&quot; iex&gt; redirect(&quot;/foo&quot;, status: 301) ...&gt; |&gt; Map.get(:status) 301 iex&gt; redirect(&quot;/foo&quot;, status: :moved_permanently) ...&gt; |&gt; Map.get(:status) 301 Notes This implementation was lifted from the sugar framework and is sufficient for many usecases. I would like to implement a back function. Complication with such functionality are discussed here - https://github.com/phoenixframework/phoenix/pull/1402 Sinatra has a very complete test suite including a back implementation - https://github.com/sinatra/sinatra/blob/9bd0d40229f76ff60d81c01ad2f4b1a8e6f31e05/test/helpers_test.rb#L183"},{"ref":"Raxx.html#request/2","title":"Raxx.request/2","type":"function","doc":"Construct a Raxx.Request. An HTTP request must have a method and path. If the location argument is a relative path the scheme and authority values will be unset. When these values can be inferred from the location they will be set. The method must be an atom for one of the HTTP methods [:GET, :POST, :PUT, :PATCH, :DELETE, :HEAD, :OPTIONS] The request will have no body or headers. These can be added with set_header/3 and set_body/2. Examples iex&gt; request(:HEAD, &quot;/&quot;).method :HEAD iex&gt; request(:GET, &quot;/&quot;).path [] iex&gt; request(:GET, &quot;/foo/bar&quot;).path [&quot;foo&quot;, &quot;bar&quot;] iex&gt; request(:GET, &quot;/foo/bar&quot;).raw_path &quot;/foo/bar&quot; iex&gt; request(:GET, &quot;https:///&quot;).scheme :https iex&gt; request(:GET, &quot;https://example.com&quot;).authority &quot;example.com&quot; iex&gt; request(:GET, &quot;/&quot;).query nil iex&gt; request(:GET, &quot;/?&quot;).query &quot;&quot; iex&gt; request(:GET, &quot;/?foo=bar&quot;).query &quot;foo=bar&quot; iex&gt; request(:GET, &quot;/&quot;).headers [] iex&gt; request(:GET, &quot;/&quot;).body false The path component of a request must contain at least / https://tools.ietf.org/html/rfc7230#section-5.3.1 If the target URI&#39;s path component is empty, the client MUST send &quot;/&quot; as the path within the origin-form of request-target. https://tools.ietf.org/html/rfc7540#section-8.1.2.3 &quot;http&quot; or &quot;https&quot; URIs that do not contain a path component MUST include a value of &#39;/&#39; iex&gt; request(:GET, &quot;&quot;).raw_path &quot;/&quot; iex&gt; request(:GET, &quot;http://example.com&quot;).raw_path &quot;/&quot;"},{"ref":"Raxx.html#request_host/1","title":"Raxx.request_host/1","type":"function","doc":"Return the host, without port, for a request. Examples iex&gt; request(:GET, &quot;http://www.example.com/hello&quot;) ...&gt; |&gt; request_host() &quot;www.example.com&quot; iex&gt; request(:GET, &quot;http://www.example.com:1234/hello&quot;) ...&gt; |&gt; request_host() &quot;www.example.com&quot;"},{"ref":"Raxx.html#request_port/1","title":"Raxx.request_port/1","type":"function","doc":"Return the host, without port, for a request. Examples iex&gt; request(:GET, &quot;http://www.example.com:1234/hello&quot;) ...&gt; |&gt; request_port() 1234 iex&gt; request(:GET, &quot;http://www.example.com/hello&quot;) ...&gt; |&gt; request_port() 80 iex&gt; request(:GET, &quot;https://www.example.com/hello&quot;) ...&gt; |&gt; request_port() 443 iex&gt; request(:GET, &quot;http://www.example.com/hello&quot;) ...&gt; |&gt; request_port(%{http: 8080}) 8080"},{"ref":"Raxx.html#request_port/2","title":"Raxx.request_port/2","type":"function","doc":"See Raxx.Request.port/2."},{"ref":"Raxx.html#response/1","title":"Raxx.response/1","type":"function","doc":"Construct a Raxx.Response. The responses HTTP status code can be provided as an integer, or will be translated from a known atom. The response will have no body or headers. These can be added with set_header/3 and set_body/2. Examples iex&gt; response(200).status 200 iex&gt; response(:no_content).status 204 iex&gt; response(200).headers [] iex&gt; response(200).body false # works for custom status codes iex&gt; response(299).status 299"},{"ref":"Raxx.html#safe?/1","title":"Raxx.safe?/1","type":"function","doc":"Request does not expect any state change on the server. For full definition of safe methods see RFC7231 Examples iex&gt; request(:GET, &quot;/&quot;) ...&gt; |&gt; safe?() true iex&gt; request(:POST, &quot;/&quot;) ...&gt; |&gt; safe?() false iex&gt; request(:PUT, &quot;/&quot;) ...&gt; |&gt; safe?() false"},{"ref":"Raxx.html#separate_parts/1","title":"Raxx.separate_parts/1","type":"function","doc":"Helper function that takes a list of Raxx.part/0 parts and breaks down Request/Response objects containing binary body values into their head, data and tail parts. All other parts get left untouched. Examples iex&gt; response = response(:ok) iex&gt; response.body false iex&gt; [response] == separate_parts([response]) true iex&gt; response_with_body = response(:ok) |&gt; set_body(&quot;some body&quot;) iex&gt; [head, data, tail] = separate_parts([response_with_body]) iex&gt; head %Raxx.Response{status: 200, body: true, headers: [{&quot;content-length&quot;, &quot;9&quot;}]} iex&gt; data %Raxx.Data{data: &quot;some body&quot;} iex&gt; tail %Raxx.Tail{headers: []}"},{"ref":"Raxx.html#set_attachment/2","title":"Raxx.set_attachment/2","type":"function","doc":"Set the &quot;content-disposition&quot; header on a response indicating the file should be downloaded. Set&#39;s the disposition to attachment the only other value of inline is assumed when no content-disposition header. NOTE: no integration with multipart downloads, see https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Content-Disposition#As_a_response_header_for_the_main_body for full details. NOTE: Sinatra sets content-type from filename extension, This is not done here in preference of being more explicit. https://github.com/sinatra/sinatra/blob/9590706ec6691520970c67b929776fd97d3c9ddd/lib/sinatra/base.rb#L362-L370 Examples iex&gt; response(:ok) ...&gt; |&gt; set_body(&quot;Hello, World!&quot;) ...&gt; |&gt; set_attachment(&quot;hello.txt&quot;) ...&gt; |&gt; get_header(&quot;content-disposition&quot;) &quot;attachment; filename=hello.txt&quot; iex&gt; response(:ok) ...&gt; |&gt; set_body(&quot;Hello, World!&quot;) ...&gt; |&gt; set_attachment(&quot;hello world.txt&quot;) ...&gt; |&gt; get_header(&quot;content-disposition&quot;) &quot;attachment; filename=hello+world.txt&quot;"},{"ref":"Raxx.html#set_body/2","title":"Raxx.set_body/2","type":"function","doc":"Add a complete body to a message. All 1xx (Informational), 204 (No Content), and 304 (Not Modified) responses do not include a message body. This functional will raise an ArgumentError if a body is set on one of these reponses. https://tools.ietf.org/html/rfc7230#section-3.1.2 Content Length When setting the body of a message to some iodata value, then the content length is also set. Examples iex&gt; request = response(:ok) ...&gt; |&gt; set_body(&quot;Hello, World!&quot;) iex&gt; request.body &quot;Hello, World!&quot; iex&gt; Raxx.get_content_length(request) 13 iex&gt; request = response(:ok) ...&gt; |&gt; set_body(true) iex&gt; request.body true iex&gt; Raxx.get_content_length(request) nil Limitations Requests using method GET or HEAD should not have a body. An HTTP GET request includes request header fields and no payload body and is therefore transmitted as a single HEADERS frame, followed by zero or more CONTINUATION frames containing the serialized block of request header fields. https://tools.ietf.org/html/rfc7540#section-8.1.3 Certain unusual usecases require a GET request with a body. For example elastic search Detailed discussion here. In such cases it is always possible to directly add the body to a request struct. Server implemenations should respect the provided body in such cases. Response with certain status codes never have a body. &gt; All 1xx (Informational), 204 (No Content), and 304 (Not Modified) responses do not include a message body. All other responses do include a message body, although the body might be of zero length. https://tools.ietf.org/html/rfc7230#section-3.3"},{"ref":"Raxx.html#set_content_length/2","title":"Raxx.set_content_length/2","type":"function","doc":"Set the content length of a request or response. The content length must be a non negative integer. Examples iex&gt; response(:ok) ...&gt; |&gt; set_content_length(13) ...&gt; |&gt; get_header(&quot;content-length&quot;) &quot;13&quot;"},{"ref":"Raxx.html#set_header/3","title":"Raxx.set_header/3","type":"function","doc":"Set the value of a header field. Examples iex&gt; request(:GET, &quot;/&quot;) ...&gt; |&gt; set_header(&quot;referer&quot;, &quot;example.com&quot;) ...&gt; |&gt; set_header(&quot;accept&quot;, &quot;text/html&quot;) ...&gt; |&gt; Map.get(:headers) [{&quot;referer&quot;, &quot;example.com&quot;}, {&quot;accept&quot;, &quot;text/html&quot;}] Limitations Raxx is protocol agnostic, i.e. it can be used to construct HTTP/1.1 or HTTP/2 messages. This limits the heads that can (or should) be set on a message The host header should not be set, this information is encoded in the authority key of a request struct. This header is forbidden on a response. A server MUST respond with a 400 (Bad Request) status code to any HTTP/1.1 request message that lacks a Host header field and to any request message that contains more than one Host header field or a Host header field with an invalid field-value. https://tools.ietf.org/html/rfc7230#section-5.4 Pseudo-header fields are only valid in the context in which they are defined. Pseudo-header fields defined for requests MUST NOT appear in responses; pseudo-header fields defined for responses MUST NOT appear in requests. https://tools.ietf.org/html/rfc7540#section-8.1.2.1 It is invalid to set a connection specific header on either a Raxx.Request or Raxx.Response. The connection specific headers are: connection keep-alive proxy-connection, transfer-encoding, upgrade Connection specific headers are not part of the end to end message, even if in HTTP/1.1 they are encoded as just another header. The &quot;Connection&quot; header field allows the sender to indicate desired control options for the current connection. In order to avoid confusing downstream recipients, a proxy or gateway MUST remove or replace any received connection options before forwarding the message. https://tools.ietf.org/html/rfc7230#section-6.1 HTTP/2 does not use the Connection header field to indicate connection-specific header fields; in this protocol, connection- specific metadata is conveyed by other means. An endpoint MUST NOT generate an HTTP/2 message containing connection-specific header fields; any message containing connection-specific header fields MUST be treated as malformed (Section 8.1.2.6) https://tools.ietf.org/html/rfc7540#section-8.1.2.2"},{"ref":"Raxx.html#set_query/2","title":"Raxx.set_query/2","type":"function","doc":"Add a query value to a request Examples iex&gt; request(:GET, &quot;/&quot;) ...&gt; |&gt; set_query(%{&quot;foo&quot; =&gt; &quot;bar&quot;}) ...&gt; |&gt; Map.get(:query) &quot;foo=bar&quot;"},{"ref":"Raxx.html#set_secure_browser_headers/1","title":"Raxx.set_secure_browser_headers/1","type":"function","doc":"Put headers that improve browser security. The following headers are set: x-frame-options - set to SAMEORIGIN to avoid clickjacking through iframes unless in the same origin x-content-type-options - set to nosniff. This requires script and style tags to be sent with proper content type x-xss-protection - set to &quot;1; mode=block&quot; to improve XSS protection on both Chrome and IE x-download-options - set to noopen to instruct the browser not to open a download directly in the browser, to avoid HTML files rendering inline and accessing the security context of the application (like critical domain cookies) x-permitted-cross-domain-policies - set to none to restrict Adobe Flash Player’s access to data Examples iex&gt; response(:ok) ...&gt; |&gt; set_secure_browser_headers() ...&gt; |&gt; get_header(&quot;x-frame-options&quot;) &quot;SAMEORIGIN&quot; iex&gt; response(:ok) ...&gt; |&gt; set_secure_browser_headers() ...&gt; |&gt; get_header(&quot;x-xss-protection&quot;) &quot;1; mode=block&quot; iex&gt; response(:ok) ...&gt; |&gt; set_secure_browser_headers() ...&gt; |&gt; get_header(&quot;x-content-type-options&quot;) &quot;nosniff&quot; iex&gt; response(:ok) ...&gt; |&gt; set_secure_browser_headers() ...&gt; |&gt; get_header(&quot;x-download-options&quot;) &quot;noopen&quot; iex&gt; response(:ok) ...&gt; |&gt; set_secure_browser_headers() ...&gt; |&gt; get_header(&quot;x-permitted-cross-domain-policies&quot;) &quot;none&quot;"},{"ref":"Raxx.html#split_path/1","title":"Raxx.split_path/1","type":"function","doc":"Split a path on forward slashes. Examples iex&gt; split_path(&quot;/foo/bar&quot;) [&quot;foo&quot;, &quot;bar&quot;]"},{"ref":"Raxx.html#tail/1","title":"Raxx.tail/1","type":"function","doc":"Construct a Raxx.Tail Examples iex&gt; tail([{&quot;digest&quot;, &quot;opaque-data&quot;}]).headers [{&quot;digest&quot;, &quot;opaque-data&quot;}] iex&gt; tail().headers []"},{"ref":"Raxx.html#t:body/0","title":"Raxx.body/0","type":"type","doc":"The body of a Raxx message. The body can be: part of the message (binary). empty (false). present but unknown (true)."},{"ref":"Raxx.html#t:header/0","title":"Raxx.header/0","type":"type","doc":"Attribute value pair that can be serialized to an HTTP request or response"},{"ref":"Raxx.html#t:headers/0","title":"Raxx.headers/0","type":"type","doc":"List of HTTP headers."},{"ref":"Raxx.html#t:message/0","title":"Raxx.message/0","type":"type","doc":"Either a Raxx.Request.t or a Raxx.Response.t"},{"ref":"Raxx.html#t:part/0","title":"Raxx.part/0","type":"type","doc":"Set of all components that make up a message to or from server."},{"ref":"Raxx.html#t:status/0","title":"Raxx.status/0","type":"type","doc":"The response status that can be parsed to e.g. response/1. This is either an integer status code like 404 or an atom that refers to a reason phrase in RFC7231 like :not_found"},{"ref":"Raxx.Context.html","title":"Raxx.Context","type":"module","doc":"Raxx.Context is a mechanism for simple sharing of state/information between Raxx.Middlewares and Raxx.Servers. It is designed to be flexible and to enable different middlewares to operate on it without conflicts. Each separate functionality using the context can be in a different &quot;section&quot;, containing arbitrary data. Context is implicitly shared using the process dictionary and persists for the duration of a single request/response cycle. If you want to pass the context to a different process, you need to take its snapshot, pass it explicitly and &quot;restore&quot; it in the other process. See Raxx.Context.get_snapshot/0 and Raxx.Context.restore_snapshot/1 for details."},{"ref":"Raxx.Context.html#delete/1","title":"Raxx.Context.delete/1","type":"function","doc":"Deletes the section from the context. Returns the previous value of the section or nil if one was not set."},{"ref":"Raxx.Context.html#get_snapshot/0","title":"Raxx.Context.get_snapshot/0","type":"function","doc":"Creates a snapshot of the current process&#39; context. The returned context data can be passed between processes and restored using Raxx.Context.restore_snapshot/1"},{"ref":"Raxx.Context.html#restore_snapshot/1","title":"Raxx.Context.restore_snapshot/1","type":"function","doc":"Restores a previously created context snapshot. It will restore the implicit state of the context for the current process to what it was when the snapshot was created using Raxx.Context.get_snapshot/0. The current context values won&#39;t be persisted in any way."},{"ref":"Raxx.Context.html#retrieve/2","title":"Raxx.Context.retrieve/2","type":"function","doc":"Retrieves the value of the context section. If the section wasn&#39;t set yet, it will return nil."},{"ref":"Raxx.Context.html#set/2","title":"Raxx.Context.set/2","type":"function","doc":"Sets the value of a context section. Returns the previous value of the section or nil if one was not set."},{"ref":"Raxx.Context.html#t:section_name/0","title":"Raxx.Context.section_name/0","type":"type","doc":""},{"ref":"Raxx.Context.html#t:snapshot/0","title":"Raxx.Context.snapshot/0","type":"opaque","doc":"An opaque type for the context snapshot data."},{"ref":"Raxx.Data.html","title":"Raxx.Data","type":"module","doc":"A part of an HTTP messages body. NOTE: There are no guarantees on how a message&#39;s body will be divided into data."},{"ref":"Raxx.Data.html#t:t/0","title":"Raxx.Data.t/0","type":"type","doc":"Container for a section of an HTTP message."},{"ref":"Raxx.HTTP1.html","title":"Raxx.HTTP1","type":"module","doc":"Toolkit for parsing and serializing requests to HTTP/1.1 format. The majority of functions return iolists and not compacted binaries. To efficiently turn a list into a binart use :erlang.iolist_to_binary/1 Notes content-length The serializer does not add the content-length header for empty bodies. The rfc7230 says it SHOULD, but there are may cases where it must not be sent. This simplifies the serialization code. It is probable that in the future Raxx.set_content_length/2 will be added. And that it will be used by Raxx.set_body/2 This is because when parsing a message the content-length headers is kept. Adding it to the Raxx.Request struct will increase the cases when serialization and deserialization result in the exact same struct. Property testing Functionality in this module might be a good opportunity for property based testing. Elixir Outlaws convinced me to give it a try. Property of serialize then decode the head should end up with the same struct Propery of any number of splits in the binary should not change the output"},{"ref":"Raxx.HTTP1.html#parse_chunk/1","title":"Raxx.HTTP1.parse_chunk/1","type":"function","doc":"Extract the content from a buffer with transfer encoding chunked"},{"ref":"Raxx.HTTP1.html#parse_request/2","title":"Raxx.HTTP1.parse_request/2","type":"function","doc":"Parse the head part of a request from a buffer. The scheme is not part of a HTTP/1.1 request, yet it is part of a HTTP/2 request. When parsing a request the scheme the buffer was received by has to be given. Options scheme (required) Set the scheme of the Raxx.Request struct. This information is not contained in the data of a HTTP/1 request. maximum_headers_count Maximum number of headers allowed in the request. maximum_line_length Maximum length (in bytes) of request line or any header line. Examples iex&gt; &quot;GET /path?qs HTTP/1.1\\r\\nhost: example.com\\r\\naccept: text/plain\\r\\n\\r\\n&quot; ...&gt; |&gt; Raxx.HTTP1.parse_request(scheme: :http) {:ok, {%Raxx.Request{ authority: &quot;example.com&quot;, body: false, headers: [{&quot;accept&quot;, &quot;text/plain&quot;}], method: :GET, path: [&quot;path&quot;], query: &quot;qs&quot;, raw_path: &quot;/path&quot;, scheme: :http }, nil, {:complete, &quot;&quot;}, &quot;&quot;}} iex&gt; &quot;GET /path?qs HTTP/1.1\\r\\nhost: example.com\\r\\naccept: text/plain\\r\\n\\r\\n&quot; ...&gt; |&gt; Raxx.HTTP1.parse_request(scheme: :https) {:ok, {%Raxx.Request{ authority: &quot;example.com&quot;, body: false, headers: [{&quot;accept&quot;, &quot;text/plain&quot;}], method: :GET, path: [&quot;path&quot;], query: &quot;qs&quot;, raw_path: &quot;/path&quot;, scheme: :https }, nil, {:complete, &quot;&quot;}, &quot;&quot;}} iex&gt; &quot;POST /path HTTP/1.1\\r\\nhost: example.com\\r\\ntransfer-encoding: chunked\\r\\ncontent-type: text/plain\\r\\n\\r\\n&quot; ...&gt; |&gt; Raxx.HTTP1.parse_request(scheme: :http) {:ok, {%Raxx.Request{ authority: &quot;example.com&quot;, body: true, headers: [{&quot;content-type&quot;, &quot;text/plain&quot;}], method: :POST, path: [&quot;path&quot;], query: nil, raw_path: &quot;/path&quot;, scheme: :http }, nil, :chunked, &quot;&quot;}} iex&gt; &quot;POST /path HTTP/1.1\\r\\nhost: example.com\\r\\ncontent-length: 13\\r\\n\\r\\n&quot; ...&gt; |&gt; Raxx.HTTP1.parse_request(scheme: :http) {:ok, {%Raxx.Request{ authority: &quot;example.com&quot;, body: true, headers: [{&quot;content-length&quot;, &quot;13&quot;}], method: :POST, path: [&quot;path&quot;], query: nil, raw_path: &quot;/path&quot;, scheme: :http }, nil, {:bytes, 13}, &quot;&quot;}} # Packet split in request line iex&gt; &quot;GET /path?qs HT&quot; ...&gt; |&gt; Raxx.HTTP1.parse_request(scheme: :http) {:more, &quot;GET /path?qs HT&quot;} # Packet split in headers iex&gt; &quot;GET / HTTP/1.1\\r\\nhost: exa&quot; ...&gt; |&gt; Raxx.HTTP1.parse_request(scheme: :http) {:more, &quot;GET / HTTP/1.1\\r\\nhost: exa&quot;} # Sending response iex&gt; &quot;HTTP/1.1 204 No Content\\r\\n&quot; ...&gt; |&gt; Raxx.HTTP1.parse_request(scheme: :http) {:error, {:invalid_line, &quot;HTTP/1.1 204 No Content\\r\\n&quot;}} # Missing host header iex&gt; &quot;GET / HTTP/1.1\\r\\naccept: text/plain\\r\\n\\r\\n&quot; ...&gt; |&gt; Raxx.HTTP1.parse_request(scheme: :http) {:error, :no_host_header} # Duplicate host header iex&gt; &quot;GET / HTTP/1.1\\r\\nhost: example.com\\r\\nhost: example2.com\\r\\n\\r\\n&quot; ...&gt; |&gt; Raxx.HTTP1.parse_request(scheme: :http) {:error, :multiple_host_headers} # Invalid content length header iex&gt; &quot;GET / HTTP/1.1\\r\\nhost: example.com\\r\\ncontent-length: eleven\\r\\n\\r\\n&quot; ...&gt; |&gt; Raxx.HTTP1.parse_request(scheme: :http) {:error, :invalid_content_length_header} # Duplicate content length header iex&gt; &quot;GET / HTTP/1.1\\r\\nhost: example.com\\r\\ncontent-length: 12\\r\\ncontent-length: 14\\r\\n\\r\\n&quot; ...&gt; |&gt; Raxx.HTTP1.parse_request(scheme: :http) {:error, :multiple_content_length_headers} # Invalid start line iex&gt; &quot;!!!BAD_REQUEST_LINE\\r\\n&quot; ...&gt; |&gt; Raxx.HTTP1.parse_request(scheme: :http) {:error, {:invalid_line, &quot;!!!BAD_REQUEST_LINE\\r\\n&quot;}} # Invalid header line iex&gt; &quot;GET / HTTP/1.1\\r\\n!!!BAD_HEADER\\r\\n\\r\\n&quot; ...&gt; |&gt; Raxx.HTTP1.parse_request(scheme: :http) {:error, {:invalid_line, &quot;!!!BAD_HEADER\\r\\n&quot;}} # Test connection status is extracted iex&gt; &quot;GET / HTTP/1.1\\r\\nhost: example.com\\r\\nconnection: close\\r\\naccept: text/plain\\r\\n\\r\\n&quot; ...&gt; |&gt; Raxx.HTTP1.parse_request(scheme: :http) {:ok, {%Raxx.Request{ authority: &quot;example.com&quot;, body: false, headers: [{&quot;accept&quot;, &quot;text/plain&quot;}], method: :GET, path: [], query: nil, raw_path: &quot;/&quot;, scheme: :http }, :close, {:complete, &quot;&quot;}, &quot;&quot;}} iex&gt; &quot;GET / HTTP/1.1\\r\\nhost: example.com\\r\\nconnection: keep-alive\\r\\naccept: text/plain\\r\\n\\r\\n&quot; ...&gt; |&gt; Raxx.HTTP1.parse_request(scheme: :http) {:ok, {%Raxx.Request{ authority: &quot;example.com&quot;, body: false, headers: [{&quot;accept&quot;, &quot;text/plain&quot;}], method: :GET, path: [], query: nil, raw_path: &quot;/&quot;, scheme: :http }, :keepalive, {:complete, &quot;&quot;}, &quot;&quot;}} # Test line_length is limited # &quot;GET /&quot; + &quot;HTTP/1.1\\r\\n&quot; = 15 iex&gt; path = &quot;/&quot; &lt;&gt; String.duplicate(&quot;a&quot;, 985) ...&gt; &quot;GET \#{path} HTTP/1.1\\r\\n&quot; ...&gt; |&gt; Raxx.HTTP1.parse_request(scheme: :http) {:error, {:line_length_limit_exceeded, :request_line}} iex&gt; path = &quot;/&quot; &lt;&gt; String.duplicate(&quot;a&quot;, 984) ...&gt; &quot;GET \#{path} HTTP/1.1\\r\\n&quot; ...&gt; |&gt; Raxx.HTTP1.parse_request(scheme: :http) ...&gt; |&gt; elem(0) :more iex&gt; path = &quot;/&quot; &lt;&gt; String.duplicate(&quot;a&quot;, 1984) ...&gt; &quot;GET \#{path} HTTP/1.1\\r\\n&quot; ...&gt; |&gt; Raxx.HTTP1.parse_request(scheme: :http, maximum_line_length: 2000) ...&gt; |&gt; elem(0) :more iex&gt; &quot;GET / HTTP/1.1\\r\\nhost: \#{String.duplicate(&quot;a&quot;, 993)}\\r\\n&quot; ...&gt; |&gt; Raxx.HTTP1.parse_request(scheme: :http) {:error, {:line_length_limit_exceeded, :header_line}} iex&gt; &quot;GET / HTTP/1.1\\r\\nhost: \#{String.duplicate(&quot;a&quot;, 992)}\\r\\n&quot; ...&gt; |&gt; Raxx.HTTP1.parse_request(scheme: :http) ...&gt; |&gt; elem(0) :more iex&gt; &quot;GET / HTTP/1.1\\r\\nhost: \#{String.duplicate(&quot;a&quot;, 1992)}\\r\\n&quot; ...&gt; |&gt; Raxx.HTTP1.parse_request(scheme: :http, maximum_line_length: 2000) ...&gt; |&gt; elem(0) :more"},{"ref":"Raxx.HTTP1.html#parse_response/2","title":"Raxx.HTTP1.parse_response/2","type":"function","doc":"Parse the head of a response. A scheme option is not given to this parser because the scheme not a requirement in HTTP/1 or HTTP/2 Options maximum_headers_count Maximum number of headers allowed in the request. maximum_line_length Maximum length (in bytes) of request line or any header line. Examples iex&gt; &quot;HTTP/1.1 204 No Content\\r\\nfoo: bar\\r\\n\\r\\n&quot; ...&gt; |&gt; Raxx.HTTP1.parse_response() {:ok, {%Raxx.Response{ status: 204, headers: [{&quot;foo&quot;, &quot;bar&quot;}], body: false }, nil, {:complete, &quot;&quot;}, &quot;&quot;}} iex&gt; &quot;HTTP/1.1 200 OK\\r\\ncontent-length: 13\\r\\ncontent-type: text/plain\\r\\n\\r\\n&quot; ...&gt; |&gt; Raxx.HTTP1.parse_response() {:ok, {%Raxx.Response{ status: 200, headers: [{&quot;content-length&quot;, &quot;13&quot;}, {&quot;content-type&quot;, &quot;text/plain&quot;}], body: true }, nil, {:bytes, 13}, &quot;&quot;}} iex&gt; &quot;HTTP/1.1 204 No Con&quot; ...&gt; |&gt; Raxx.HTTP1.parse_response() {:more, :undefined} iex&gt; &quot;HTTP/1.1 204 No Content\\r\\nfo&quot; ...&gt; |&gt; Raxx.HTTP1.parse_response() {:more, :undefined} # Request given iex&gt; &quot;GET / HTTP/1.1\\r\\n&quot; ...&gt; |&gt; Raxx.HTTP1.parse_response() {:error, {:invalid_line, &quot;GET / HTTP/1.1\\r\\n&quot;}} iex&gt; &quot;!!!BAD_STATUS_LINE\\r\\n&quot; ...&gt; |&gt; Raxx.HTTP1.parse_response() {:error, {:invalid_line, &quot;!!!BAD_STATUS_LINE\\r\\n&quot;}} iex&gt; &quot;HTTP/1.1 204 No Content\\r\\n!!!BAD_HEADER\\r\\n\\r\\n&quot; ...&gt; |&gt; Raxx.HTTP1.parse_response() {:error, {:invalid_line, &quot;!!!BAD_HEADER\\r\\n&quot;}} iex&gt; &quot;HTTP/1.1 204 No Content\\r\\nconnection: close\\r\\nfoo: bar\\r\\n\\r\\n&quot; ...&gt; |&gt; Raxx.HTTP1.parse_response() {:ok, {%Raxx.Response{ status: 204, headers: [{&quot;foo&quot;, &quot;bar&quot;}], body: false }, :close, {:complete, &quot;&quot;}, &quot;&quot;}} iex&gt; &quot;HTTP/1.1 204 No Content\\r\\nconnection: keep-alive\\r\\nfoo: bar\\r\\n\\r\\n&quot; ...&gt; |&gt; Raxx.HTTP1.parse_response() {:ok, {%Raxx.Response{ status: 204, headers: [{&quot;foo&quot;, &quot;bar&quot;}], body: false }, :keepalive, {:complete, &quot;&quot;}, &quot;&quot;}} # Test exceptional case when server returns Title case values iex&gt; &quot;HTTP/1.1 204 No Content\\r\\nconnection: Close\\r\\nfoo: bar\\r\\n\\r\\n&quot; ...&gt; |&gt; Raxx.HTTP1.parse_response() {:ok, {%Raxx.Response{ status: 204, headers: [{&quot;foo&quot;, &quot;bar&quot;}], body: false }, :close, {:complete, &quot;&quot;}, &quot;&quot;}} # Test exceptional case when server uses Title case values iex&gt; &quot;HTTP/1.1 204 No Content\\r\\nconnection: Keep-alive\\r\\nfoo: bar\\r\\n\\r\\n&quot; ...&gt; |&gt; Raxx.HTTP1.parse_response() {:ok, {%Raxx.Response{ status: 204, headers: [{&quot;foo&quot;, &quot;bar&quot;}], body: false }, :keepalive, {:complete, &quot;&quot;}, &quot;&quot;}} # Invalid connection header iex&gt; &quot;HTTP/1.1 204 No Content\\r\\nconnection: Invalid\\r\\n\\r\\n&quot; ...&gt; |&gt; Raxx.HTTP1.parse_response() {:error, :invalid_connection_header} # duplicate connection header iex&gt; &quot;HTTP/1.1 204 No Content\\r\\nconnection: close\\r\\nconnection: keep-alive\\r\\n\\r\\n&quot; ...&gt; |&gt; Raxx.HTTP1.parse_response() {:error, :multiple_connection_headers} # Test line_length is limited # &quot;HTTP/1.1 204 &quot; + newlines = 15 iex&gt; reason_phrase = String.duplicate(&quot;A&quot;, 986) ...&gt; &quot;HTTP/1.1 204 \#{reason_phrase}\\r\\n&quot; ...&gt; |&gt; Raxx.HTTP1.parse_response() {:error, {:line_length_limit_exceeded, :status_line}} iex&gt; reason_phrase = String.duplicate(&quot;A&quot;, 985) ...&gt; &quot;HTTP/1.1 204 \#{reason_phrase}\\r\\n&quot; ...&gt; |&gt; Raxx.HTTP1.parse_response() ...&gt; |&gt; elem(0) :more iex&gt; reason_phrase = String.duplicate(&quot;A&quot;, 1985) ...&gt; &quot;HTTP/1.1 204 \#{reason_phrase}\\r\\n&quot; ...&gt; |&gt; Raxx.HTTP1.parse_response(maximum_line_length: 2000) ...&gt; |&gt; elem(0) :more iex&gt; &quot;HTTP/1.1 204 No Content\\r\\nfoo: \#{String.duplicate(&quot;a&quot;, 994)}\\r\\n&quot; ...&gt; |&gt; Raxx.HTTP1.parse_response() {:error, {:line_length_limit_exceeded, :header_line}} iex&gt; &quot;HTTP/1.1 204 No Content\\r\\nfoo: \#{String.duplicate(&quot;a&quot;, 993)}\\r\\n&quot; ...&gt; |&gt; Raxx.HTTP1.parse_response() ...&gt; |&gt; elem(0) :more iex&gt; &quot;HTTP/1.1 204 No Content\\r\\nfoo: \#{String.duplicate(&quot;a&quot;, 1993)}\\r\\n&quot; ...&gt; |&gt; Raxx.HTTP1.parse_response(maximum_line_length: 2000) ...&gt; |&gt; elem(0) :more # Test maximum number of headers is limited iex&gt; &quot;HTTP/1.1 204 No Content\\r\\n\#{String.duplicate(&quot;foo: bar\\r\\n&quot;, 101)}&quot; ...&gt; |&gt; Raxx.HTTP1.parse_response() {:error, {:header_count_exceeded, 100}} # Test maximum number of headers is limited iex&gt; &quot;HTTP/1.1 204 No Content\\r\\n\#{String.duplicate(&quot;foo: bar\\r\\n&quot;, 2)}&quot; ...&gt; |&gt; Raxx.HTTP1.parse_response(maximum_headers_count: 1) {:error, {:header_count_exceeded, 1}}"},{"ref":"Raxx.HTTP1.html#serialize_chunk/1","title":"Raxx.HTTP1.serialize_chunk/1","type":"function","doc":"Serialize io_data as a single chunk to be streamed. Example iex&gt; Raxx.HTTP1.serialize_chunk(&quot;hello&quot;) ...&gt; |&gt; to_string() &quot;5\\r\\nhello\\r\\n&quot; iex&gt; Raxx.HTTP1.serialize_chunk(&quot;&quot;) ...&gt; |&gt; to_string() &quot;0\\r\\n\\r\\n&quot;"},{"ref":"Raxx.HTTP1.html#serialize_request/2","title":"Raxx.HTTP1.serialize_request/2","type":"function","doc":"Serialize a request to wire format NOTE set_body should add content-length otherwise we don&#39;t know if to delete it to match on other end, when serializing https://tools.ietf.org/html/rfc7230#section-5.4 Since the Host field-value is critical information for handling a request, a user agent SHOULD generate Host as the first header field following the request-line. Examples iex&gt; request = Raxx.request(:GET, &quot;http://example.com/path?qs&quot;) ...&gt; |&gt; Raxx.set_header(&quot;accept&quot;, &quot;text/plain&quot;) ...&gt; {head, body} = Raxx.HTTP1.serialize_request(request) ...&gt; :erlang.iolist_to_binary(head) &quot;GET /path?qs HTTP/1.1\\r\\nhost: example.com\\r\\naccept: text/plain\\r\\n\\r\\n&quot; iex&gt; body {:complete, &quot;&quot;} iex&gt; request = Raxx.request(:POST, &quot;https://example.com&quot;) ...&gt; |&gt; Raxx.set_header(&quot;content-type&quot;, &quot;text/plain&quot;) ...&gt; |&gt; Raxx.set_body(true) ...&gt; {head, body} = Raxx.HTTP1.serialize_request(request) ...&gt; :erlang.iolist_to_binary(head) &quot;POST / HTTP/1.1\\r\\nhost: example.com\\r\\ntransfer-encoding: chunked\\r\\ncontent-type: text/plain\\r\\n\\r\\n&quot; iex&gt; body :chunked iex&gt; request = Raxx.request(:POST, &quot;https://example.com&quot;) ...&gt; |&gt; Raxx.set_header(&quot;content-length&quot;, &quot;13&quot;) ...&gt; |&gt; Raxx.set_body(true) ...&gt; {head, body} = Raxx.HTTP1.serialize_request(request) ...&gt; :erlang.iolist_to_binary(head) &quot;POST / HTTP/1.1\\r\\nhost: example.com\\r\\ncontent-length: 13\\r\\n\\r\\n&quot; iex&gt; body {:bytes, 13} https://tools.ietf.org/html/rfc7230#section-6.1 A client that does not support persistent connections MUST send the &quot;close&quot; connection option in every request message. iex&gt; request = Raxx.request(:GET, &quot;http://example.com/&quot;) ...&gt; |&gt; Raxx.set_header(&quot;accept&quot;, &quot;text/plain&quot;) ...&gt; {head, _body} = Raxx.HTTP1.serialize_request(request, connection: :close) ...&gt; :erlang.iolist_to_binary(head) &quot;GET / HTTP/1.1\\r\\nhost: example.com\\r\\nconnection: close\\r\\naccept: text/plain\\r\\n\\r\\n&quot; iex&gt; request = Raxx.request(:GET, &quot;http://example.com/&quot;) ...&gt; |&gt; Raxx.set_header(&quot;accept&quot;, &quot;text/plain&quot;) ...&gt; {head, _body} = Raxx.HTTP1.serialize_request(request, connection: :keepalive) ...&gt; :erlang.iolist_to_binary(head) &quot;GET / HTTP/1.1\\r\\nhost: example.com\\r\\nconnection: keep-alive\\r\\naccept: text/plain\\r\\n\\r\\n&quot;"},{"ref":"Raxx.HTTP1.html#serialize_response/2","title":"Raxx.HTTP1.serialize_response/2","type":"function","doc":"Serialize a response to an iolist Because of HEAD requests we should keep body separate Examples iex&gt; response = Raxx.response(200) ...&gt; |&gt; Raxx.set_header(&quot;content-type&quot;, &quot;text/plain&quot;) ...&gt; |&gt; Raxx.set_body(&quot;Hello, World!&quot;) ...&gt; {head, body} = Raxx.HTTP1.serialize_response(response) ...&gt; :erlang.iolist_to_binary(head) &quot;HTTP/1.1 200 OK\\r\\ncontent-type: text/plain\\r\\ncontent-length: 13\\r\\n\\r\\n&quot; iex&gt; body {:complete, &quot;Hello, World!&quot;} iex&gt; response = Raxx.response(200) ...&gt; |&gt; Raxx.set_header(&quot;content-length&quot;, &quot;13&quot;) ...&gt; |&gt; Raxx.set_header(&quot;content-type&quot;, &quot;text/plain&quot;) ...&gt; |&gt; Raxx.set_body(true) ...&gt; {head, body} = Raxx.HTTP1.serialize_response(response) ...&gt; :erlang.iolist_to_binary(head) &quot;HTTP/1.1 200 OK\\r\\ncontent-length: 13\\r\\ncontent-type: text/plain\\r\\n\\r\\n&quot; iex&gt; body {:bytes, 13} iex&gt; response = Raxx.response(200) ...&gt; |&gt; Raxx.set_header(&quot;content-type&quot;, &quot;text/plain&quot;) ...&gt; |&gt; Raxx.set_body(true) ...&gt; {head, body} = Raxx.HTTP1.serialize_response(response) ...&gt; :erlang.iolist_to_binary(head) &quot;HTTP/1.1 200 OK\\r\\ntransfer-encoding: chunked\\r\\ncontent-type: text/plain\\r\\n\\r\\n&quot; iex&gt; body :chunked &gt; A server MUST NOT send a Content-Length header field in any response &gt; with a status code of 1xx (Informational) or 204 (No Content). A &gt; server MUST NOT send a Content-Length header field in any 2xx &gt; (Successful) response to a CONNECT request (Section 4.3.6 of &gt; [RFC7231]). iex&gt; Raxx.response(204) ...&gt; |&gt; Raxx.set_header(&quot;foo&quot;, &quot;bar&quot;) ...&gt; |&gt; Raxx.HTTP1.serialize_response() ...&gt; |&gt; elem(0) ...&gt; |&gt; :erlang.iolist_to_binary() &quot;HTTP/1.1 204 No Content\\r\\nfoo: bar\\r\\n\\r\\n&quot; https://tools.ietf.org/html/rfc7230#section-6.1 A server that does not support persistent connections MUST send the &quot;close&quot; connection option in every response message that does not have a 1xx (Informational) status code. iex&gt; Raxx.response(204) ...&gt; |&gt; Raxx.set_header(&quot;foo&quot;, &quot;bar&quot;) ...&gt; |&gt; Raxx.HTTP1.serialize_response(connection: :close) ...&gt; |&gt; elem(0) ...&gt; |&gt; :erlang.iolist_to_binary() &quot;HTTP/1.1 204 No Content\\r\\nconnection: close\\r\\nfoo: bar\\r\\n\\r\\n&quot; iex&gt; Raxx.response(204) ...&gt; |&gt; Raxx.set_header(&quot;foo&quot;, &quot;bar&quot;) ...&gt; |&gt; Raxx.HTTP1.serialize_response(connection: :keepalive) ...&gt; |&gt; elem(0) ...&gt; |&gt; :erlang.iolist_to_binary() &quot;HTTP/1.1 204 No Content\\r\\nconnection: keep-alive\\r\\nfoo: bar\\r\\n\\r\\n&quot;"},{"ref":"Raxx.HTTP1.html#t:body_read_state/0","title":"Raxx.HTTP1.body_read_state/0","type":"type","doc":""},{"ref":"Raxx.HTTP1.html#t:connection_status/0","title":"Raxx.HTTP1.connection_status/0","type":"type","doc":""},{"ref":"Raxx.Middleware.html","title":"Raxx.Middleware","type":"behaviour","doc":"A &quot;middleware&quot; is a component that sits between the HTTP server such as as Ace and a Raxx.Server controller. The middleware can modify requests request before giving it to the controller and modify the controllers response before it&#39;s given to the server. Oftentimes multiple middlewaress might be attached to a controller and function as a single Raxx.Server.t/0 - see Raxx.Stack for details. The Raxx.Middleware provides a behaviour to be implemented by middlewares. Example Traditionally, middlewares are used for a variety of purposes: managing CORS, CSRF protection, logging, error handling, and many more. This example shows a middleware that given a HEAD request &quot;translates&quot; it to a GET one, hands it over to the controller and strips the response body transforms the response according to RFC 2616 This way the controller doesn&#39;t heed to handle the HEAD case at all. defmodule Raxx.Middleware.Head do alias Raxx.Server alias Raxx.Middleware @behaviour Middleware @impl Middleware def process_head(request = %{method: :HEAD}, _config, inner_server) do request = %{request | method: :GET} state = :engage {parts, inner_server} = Server.handle_head(inner_server, request) parts = modify_response_parts(parts, state) {parts, state, inner_server} end def process_head(request = %{method: _}, _config, inner_server) do {parts, inner_server} = Server.handle_head(inner_server, request) {parts, :disengage, inner_server} end @impl Middleware def process_data(data, state, inner_server) do {parts, inner_server} = Server.handle_data(inner_server, data) parts = modify_response_parts(parts, state) {parts, state, inner_server} end @impl Middleware def process_tail(tail, state, inner_server) do {parts, inner_server} = Server.handle_tail(inner_server, tail) parts = modify_response_parts(parts, state) {parts, state, inner_server} end @impl Middleware def process_info(info, state, inner_server) do {parts, inner_server} = Server.handle_info(inner_server, info) parts = modify_response_parts(parts, state) {parts, state, inner_server} end defp modify_response_parts(parts, :disengage) do parts end defp modify_response_parts(parts, :engage) do Enum.flat_map(parts, &amp;do_handle_response_part(&amp;1)) end defp do_handle_response_part(response = %Raxx.Response{}) do # the content-length will remain the same [%Raxx.Response{response | body: false}] end defp do_handle_response_part(%Raxx.Data{}) do [] end defp do_handle_response_part(%Raxx.Tail{}) do [] end end Within the callback implementations the middleware should call through to the &quot;inner&quot; server and make sure to return its updated state as part of the Raxx.Middleware.next/0 tuple. In certain situations the middleware might want to short-circuit processing of the incoming messages, bypassing the server. In that case, it should not call through using Raxx.Server&#39;s handle_* helper functions and return the inner_server unmodified. NOTE: As you can see in the above example, the middleware can even modify the info messages sent to the server and is responsible for forwarding them."},{"ref":"Raxx.Middleware.html#c:process_data/3","title":"Raxx.Middleware.process_data/3","type":"callback","doc":"Called every time data from the request body is received."},{"ref":"Raxx.Middleware.html#c:process_head/3","title":"Raxx.Middleware.process_head/3","type":"callback","doc":"Called once when a client starts a stream, The arguments a Raxx.Request, the middleware configuration and the &quot;inner&quot; server for the middleware to call through to. This callback can be relied upon to execute before any other callbacks"},{"ref":"Raxx.Middleware.html#c:process_info/3","title":"Raxx.Middleware.process_info/3","type":"callback","doc":"Called for all other messages the middleware may recieve. The middleware is responsible for forwarding them to the inner server."},{"ref":"Raxx.Middleware.html#c:process_tail/3","title":"Raxx.Middleware.process_tail/3","type":"callback","doc":"Called once when a request finishes. This will be called with an empty list of headers is request is completed without trailers. Will not be called at all if the Raxx.Request.t/0 passed to process_head/3 had body: false."},{"ref":"Raxx.Middleware.html#t:next/0","title":"Raxx.Middleware.next/0","type":"type","doc":"Values returned from the process_* callbacks"},{"ref":"Raxx.Middleware.html#t:state/0","title":"Raxx.Middleware.state/0","type":"type","doc":"State of middleware."},{"ref":"Raxx.Middleware.html#t:t/0","title":"Raxx.Middleware.t/0","type":"type","doc":"The behaviour module and state/config of a raxx middleware"},{"ref":"Raxx.Request.html","title":"Raxx.Request","type":"module","doc":"HTTP requests to a Raxx application are encapsulated in a Raxx.Request struct. A request has all the properties of the url it was sent to. In addition it has optional content, in the body. As well as a variable number of headers that contain meta data. Where appropriate URI properties are named from this definition. scheme:[//[user:password@]host[:port]][/]path[?query][#fragment] from wikipedia The contents are itemised below: schemehttp or https, depending on the transport used. authorityThe location of the hosting server, as a binary. e.g. www.example.com. Plus an optional port number, separated from the hostname by a colon methodThe HTTP request method, such as :GET or :POST, as an atom. This cannot ever be nil. It is always uppercase. pathThe remainder of the request URL&#39;s “path”, split into segments. It designates the virtual “location” of the request&#39;s target within the application. This may be an empty array, if the requested URL targets the application root. raw_pathThe request URL&#39;s &quot;path&quot; querythe URL query string. headersThe headers from the HTTP request as a proplist of strings. Note all headers will be downcased, e.g. [{&quot;content-type&quot;, &quot;text/plain&quot;}] bodyThe body content sent with the request"},{"ref":"Raxx.Request.html#host/1","title":"Raxx.Request.host/1","type":"function","doc":""},{"ref":"Raxx.Request.html#port/2","title":"Raxx.Request.port/2","type":"function","doc":""},{"ref":"Raxx.Request.html#t:method/0","title":"Raxx.Request.method/0","type":"type","doc":"Method to indicate the desired action to be performed on the identified resource."},{"ref":"Raxx.Request.html#t:scheme/0","title":"Raxx.Request.scheme/0","type":"type","doc":"Scheme describing protocol used."},{"ref":"Raxx.Request.html#t:t/0","title":"Raxx.Request.t/0","type":"type","doc":"Elixir representation for an HTTP request."},{"ref":"Raxx.Response.html","title":"Raxx.Response","type":"module","doc":"HTTP responses from a Raxx application are encapsulated in a Raxx.Response struct. The contents are itemised below: statusThe HTTP status code for the response: 1xx, 2xx, 3xx, 4xx, 5xx headersThe response headers as a list: [{&quot;content-type&quot;, &quot;text/plain&quot;} bodyThe response body, by default an empty string."},{"ref":"Raxx.Response.html#t:status_code/0","title":"Raxx.Response.status_code/0","type":"type","doc":"Integer code for server response type"},{"ref":"Raxx.Response.html#t:t/0","title":"Raxx.Response.t/0","type":"type","doc":"Elixir representation for an HTTP response."},{"ref":"Raxx.Router.html","title":"Raxx.Router","type":"behaviour","doc":"Routing for Raxx applications. Routes are defined as a match and an action module. Standard Elixir pattern matching is used to apply the match to an incoming request. An action module another implementation of Raxx.Server Sections group routes that all have the same middleware. Middleware in a section maybe defined as a list, this is useful when all configuration is known at compile-time. Alternativly an arity 1 function can be used. This can be used when middleware require runtime configuration. The argument passed to this function is server initial state. Examples defmodule MyRouter do use Raxx.Router section [{Raxx.Logger, level: :debug}], [ {%{method: :GET, path: [&quot;ping&quot;]}, Ping}, ] section &amp;web/1, [ {%{method: :GET, path: []}, HomePage}, {%{method: :GET, path: [&quot;users&quot;]}, UsersPage}, {%{method: :GET, path: [&quot;users&quot;, _id]}, UserPage}, {%{method: :POST, path: [&quot;users&quot;]}, CreateUser}, {_, NotFoundPage} ] def web(state) do [ {Raxx.Logger, level: state.log_level}, {MyMiddleware, foo: state.foo} ] end end If the sections DSL does not work for an application it is possible to instead just implement a route/2 function."},{"ref":"Raxx.Router.html#c:route/2","title":"Raxx.Router.route/2","type":"callback","doc":""},{"ref":"Raxx.Router.html#section/2","title":"Raxx.Router.section/2","type":"macro","doc":"Define a set of routes with a common set of middlewares applied to them. The first argument may be a list of middlewares; or a function that accepts one argument, the initial state, and returns a list of middleware. If all settings for a middleware can be decided at compile-time then a list is preferable."},{"ref":"Raxx.Server.html","title":"Raxx.Server","type":"behaviour","doc":"Interface to handle server side communication in an HTTP message exchange. If simple request -&gt; response transformation is possible, try Raxx.SimpleServer A module implementing Raxx.Server is run by an HTTP server. For example Ace can run such a module for both HTTP/1.x and HTTP/2 exchanges Getting Started Send complete response as soon as request headers are received. defmodule SimpleServer do use Raxx.Server def handle_head(%Raxx.Request{method: :GET, path: []}, _state) do response(:ok) |&gt; set_header(&quot;content-type&quot;, &quot;text/plain&quot;) |&gt; set_body(&quot;Hello, World!&quot;) end end Store data as it is available from a clients request defmodule StreamingRequest do use Raxx.Server def handle_head(%Raxx.Request{method: :PUT, body: true}, _state) do {:ok, io_device} = File.open(&quot;my/path&quot;) {[], {:file, device}} end def handle_data(body, state = {:file, device}) do IO.write(device, body) {[], state} end def handle_tail(_trailers, state) do response(:see_other) |&gt; set_header(&quot;location&quot;, &quot;/&quot;) end end Subscribe server to event source and forward notifications to client. defmodule SubscribeToMessages do use Raxx.Server def handle_head(_request, _state) do {:ok, _} = ChatRoom.join() response(:ok) |&gt; set_header(&quot;content-type&quot;, &quot;text/event-stream&quot;) |&gt; set_body(true) end def handle_info({ChatRoom, data}, state) do {[body(data)], state} end end Notes handle_head/2 will always be called with a request that has body as a boolean. For small requests where buffering the whole request is acceptable a simple middleware can be used. Acceptable return values are the same for all callbacks; either a Raxx.Response, which must be complete or a list of message parts and a new state. Streaming Raxx.Server defines an interface to stream the body of request and responses. This has several advantages: Large payloads do not need to be help in memory Server can push information as it becomes available, using Server Sent Events. If a request has invalid headers then a reply can be set without handling the body. Content can be generated as requested using HTTP/2 flow control The body of a Raxx message (Raxx.Request or Raxx.Response) may be one of three types: io_list - This is the complete body for the message. :false - There is no body, for example :GET requests never have a body. :true - There is a body, it can be processed as it is received Server Isolation To start an exchange a client sends a request. The server, upon receiving this message, sends a reply. A logical HTTP exchange consists of a single request and response. Methods such as pipelining and multiplexing combine multiple logical exchanges onto a single connection. This is done to improve performance and is a detail not exposed a server. A Raxx server handles a single HTTP exchange. Therefore a single connection my have multiple servers each isolated in their own process. Termination An exchange can be stopped early by terminating the server process. Support for early termination is not consistent between versions of HTTP. HTTP/2: server exit with reason :normal, stream reset with error CANCEL. HTTP/2: server exit any other reason, stream reset with error INTERNAL_ERROR. HTTP/1.x: server exit with any reason, connection is closed. Raxx.Server does not provide a terminate callback. Any cleanup that needs to be done from an aborted exchange should be handled by monitoring the server process."},{"ref":"Raxx.Server.html#handle/2","title":"Raxx.Server.handle/2","type":"function","doc":"Execute a server module and current state in response to a new message"},{"ref":"Raxx.Server.html#handle_data/2","title":"Raxx.Server.handle_data/2","type":"function","doc":"Similar to Raxx.Server.handle/2, except it only accepts the &quot;unpacked&quot;, binary data and returns the whole server, not just its state. Raxx.Server.handle/2 uses the data structures sent from the http server, whereas Raxx.Server.handle_* use the &quot;unpacked&quot; data, in the shape defined by callbacks."},{"ref":"Raxx.Server.html#c:handle_data/2","title":"Raxx.Server.handle_data/2","type":"callback","doc":"Called every time data from the request body is received"},{"ref":"Raxx.Server.html#handle_head/2","title":"Raxx.Server.handle_head/2","type":"function","doc":"Similar to Raxx.Server.handle/2, except it only accepts Raxx.Request.t/0 and returns the whole server, not just its state. Raxx.Server.handle/2 uses the data structures sent from the http server, whereas Raxx.Server.handle_* use the &quot;unpacked&quot; data, in the shape defined by callbacks."},{"ref":"Raxx.Server.html#c:handle_head/2","title":"Raxx.Server.handle_head/2","type":"callback","doc":"Called once when a client starts a stream, Passed a Raxx.Request and server configuration. Note the value of the request body will be a boolean. This callback can be relied upon to execute before any other callbacks"},{"ref":"Raxx.Server.html#handle_info/2","title":"Raxx.Server.handle_info/2","type":"function","doc":"Similar to Raxx.Server.handle/2, except it only accepts the &quot;unpacked&quot;, trailers and returns the whole server, not just its state. Raxx.Server.handle/2 uses the data structures sent from the http server, whereas Raxx.Server.handle_* use the &quot;unpacked&quot; data, in the shape defined by callbacks."},{"ref":"Raxx.Server.html#c:handle_info/2","title":"Raxx.Server.handle_info/2","type":"callback","doc":"Called for all other messages the server may recieve"},{"ref":"Raxx.Server.html#handle_tail/2","title":"Raxx.Server.handle_tail/2","type":"function","doc":"Similar to Raxx.Server.handle/2, except it only accepts the &quot;unpacked&quot;, trailers and returns the whole server, not just its state. Raxx.Server.handle/2 uses the data structures sent from the http server, whereas Raxx.Server.handle_* use the &quot;unpacked&quot; data, in the shape defined by callbacks."},{"ref":"Raxx.Server.html#c:handle_tail/2","title":"Raxx.Server.handle_tail/2","type":"callback","doc":"Called once when a request finishes. This will be called with an empty list of headers is request is completed without trailers. Will not be called at all if the Raxx.Request.t/0 struct passed to handle_head/2 had body: false."},{"ref":"Raxx.Server.html#verify_server/1","title":"Raxx.Server.verify_server/1","type":"function","doc":"Verify server can be run? A runnable server consists of a tuple of server module and initial state. The server module must implement this modules behaviour. The initial state can be any term Examples # Could just call verify iex&gt; Raxx.Server.verify_server({Raxx.ServerTest.DefaultServer, %{}}) {:ok, {Raxx.ServerTest.DefaultServer, %{}}} iex&gt; Raxx.Server.verify_server({GenServer, %{}}) {:error, {:not_a_server_module, GenServer}} iex&gt; Raxx.Server.verify_server({NotAModule, %{}}) {:error, {:not_a_module, NotAModule}}"},{"ref":"Raxx.Server.html#t:next/0","title":"Raxx.Server.next/0","type":"type","doc":"Possible return values instructing server to send client data and update state if appropriate."},{"ref":"Raxx.Server.html#t:state/0","title":"Raxx.Server.state/0","type":"type","doc":"State of application server. Original value is the configuration given when starting the raxx application."},{"ref":"Raxx.Server.html#t:t/0","title":"Raxx.Server.t/0","type":"type","doc":"The behaviour and state of a raxx server"},{"ref":"Raxx.SimpleServer.html","title":"Raxx.SimpleServer","type":"behaviour","doc":"Server interface for simple request -&gt; response interactions. Modules that use Raxx.SimpleServer implement the Raxx.Server behaviour. Default implementations are provided for the streaming interface to buffer the request before a single call to handle_request/2. Example Echo the body of a request to the client defmodule EchoServer do use Raxx.SimpleServer, maximum_body_length: 12 * 1024 * 1024 def handle_request(%Raxx.Request{method: :POST, path: [], body: body}, _state) do response(:ok) |&gt; set_header(&quot;content-type&quot;, &quot;text/plain&quot;) |&gt; set_body(body) end end Options maximum_body_length (default 8MB) the maximum sized body that will be automatically buffered. For large requests, e.g. file uploads, consider implementing a streaming server."},{"ref":"Raxx.SimpleServer.html#c:handle_request/2","title":"Raxx.SimpleServer.handle_request/2","type":"callback","doc":"Called with a complete request once all the data parts of a body are received. Passed a Raxx.Request and server configuration. Note the value of the request body will be a string."},{"ref":"Raxx.SimpleServer.html#t:state/0","title":"Raxx.SimpleServer.state/0","type":"type","doc":"State of application server. Original value is the configuration given when starting the raxx application."},{"ref":"Raxx.Stack.html","title":"Raxx.Stack","type":"module","doc":"A Raxx.Stack is a list of Raxx.Middlewares attached to a Raxx.Server. It implements the Raxx.Server interface itself so it can be used anywhere &quot;normal&quot; server can be."},{"ref":"Raxx.Stack.html#get_middlewares/1","title":"Raxx.Stack.get_middlewares/1","type":"function","doc":"Returns the server contained in the stack."},{"ref":"Raxx.Stack.html#get_server/1","title":"Raxx.Stack.get_server/1","type":"function","doc":"Returns the server contained in the stack."},{"ref":"Raxx.Stack.html#new/2","title":"Raxx.Stack.new/2","type":"function","doc":"Creates a new stack from a list of middlewares and a server."},{"ref":"Raxx.Stack.html#set_middlewares/2","title":"Raxx.Stack.set_middlewares/2","type":"function","doc":"Replaces the middlewares in the stack."},{"ref":"Raxx.Stack.html#set_server/2","title":"Raxx.Stack.set_server/2","type":"function","doc":"Replaces the server in the stack."},{"ref":"Raxx.Stack.html#t:state/0","title":"Raxx.Stack.state/0","type":"opaque","doc":"The internal state of the Raxx.Stack. Its structure shouldn&#39;t be relied on, it is subject to change without warning."},{"ref":"Raxx.Stack.html#t:t/0","title":"Raxx.Stack.t/0","type":"type","doc":"Represents a pipeline of middlewares attached to a server. Can be used exactly as any Raxx.Server.t/0 could be."},{"ref":"Raxx.Tail.html","title":"Raxx.Tail","type":"module","doc":"A trailer allows the sender to include additional fields at the end of a streamed message."},{"ref":"Raxx.Tail.html#t:t/0","title":"Raxx.Tail.t/0","type":"type","doc":"Container for optional trailers of an HTTP message."},{"ref":"ReturnError.html","title":"ReturnError","type":"exception","doc":"Raise when a server module returns an invalid reaction"},{"ref":"ReturnError.html#message/1","title":"ReturnError.message/1","type":"function","doc":"Callback implementation for Exception.message/1."},{"ref":"readme.html","title":"Raxx","type":"extras","doc":"Raxx Interface for HTTP webservers, frameworks and clients. Install from hex.pm Documentation available on hexdoc Discuss on slack See Raxx.Kit for a project generator that helps you set up a web project based on Raxx/Ace."},{"ref":"readme.html#simple-server","title":"Raxx - Simple server","type":"extras","doc":"1. Defining a server defmodule MyServer do use Raxx.SimpleServer @impl Raxx.SimpleServer def handle_request(%{method: :GET, path: []}, _state) do response(:ok) |&gt; set_header(&quot;content-type&quot;, &quot;text/plain&quot;) |&gt; set_body(&quot;Hello, World!&quot;) end def handle_request(%{method: :GET, path: _}, _state) do response(:not_found) |&gt; set_header(&quot;content-type&quot;, &quot;text/plain&quot;) |&gt; set_body(&quot;Oops! Nothing here.&quot;) end end A request&#39;s path is split into segments. A request to GET / has path []. 2. Running a server To start a Raxx server a compatible HTTP server is needed. This example uses Ace that can serve both HTTP/1 and HTTP/2. raxx_server = {MyServer, nil} http_options = [port: 8080, cleartext: true] {:ok, pid} = Ace.HTTP.Service.start_link(raxx_server, http_options) The second element in the Raxx server tuple is passed as the second argument to the handle_request/2 callback. In this example it is unused and so set to nil. Start your project and visit http://localhost:8080."},{"ref":"readme.html#http-streaming","title":"Raxx - HTTP streaming","type":"extras","doc":"An HTTP exchange involves a client sending data to a server receiving a response. A simple view is to model this as a single message sent in each direction. Working with this model corresponds to Raxx.SimpleServer callbacks. request --&gt; Client ============================================ Server &lt;-- response When the simple model is insufficient Raxx exposes a lower model. This consists of a series of messages in each direction. Working with this model corresponds to Raxx.Server callbacks. tail | data(1+) | head(request) --&gt; Client ============================================ Server &lt;-- head(response) | data(1+) | tail The body of a request or a response, is the combination of all data parts sent. Stateful server The LongPoll server is stateful. After receiving a complete request this server has to wait for extra input before sending a response to the client. defmodule LongPoll do use Raxx.Server @impl Raxx.Server def handle_head(%{method: :GET, path: [&quot;slow&quot;]}, state) do Process.send_after(self(), :reply, 30_000) {[], state} end @impl Raxx.Server def handle_info(:reply, _state) do response(:ok) |&gt; set_header(&quot;content-type&quot;, &quot;text/plain&quot;) |&gt; set_body(&quot;Hello, Thanks for waiting.&quot;) end end A long lived server needs to return two things; the message parts to send, in this case nothing []; and the new state of the server, in this case no change state. The initial_state is configured when the server is started. Server streaming The SubscribeToMessages server streams its response. The server will send the head of the response upon receiving the request. Data is sent to the client, as part of the body, when it becomes available. The response is completed when the chatroom sends a :closed message. defmodule SubscribeToMessages do use Raxx.Server @impl Raxx.Server def handle_head(%{method: :GET, path: [&quot;messages&quot;]}, state) do {:ok, _} = ChatRoom.join() outbound = response(:ok) |&gt; set_header(&quot;content-type&quot;, &quot;text/plain&quot;) |&gt; set_body(true) {[outbound], state} end @impl Raxx.Server def handle_info({ChatRoom, :closed}, state) do outbound = tail() {[outbound], state} end def handle_info({ChatRoom, data}, state) do outbound = data(data) {[outbound], state} end end Using set_body(true) marks that the response has a body that it is not yet known. A stream must have a tail to complete, metadata added here will be sent as trailers. Client streaming The Upload server writes data to a file as it is received. Only once the complete request has been received is a response sent. defmodule Upload do use Raxx.Server @impl Raxx.Server def handle_head(%{method: :PUT, path: [&quot;upload&quot;] body: true}, _state) do {:ok, io_device} = File.open(&quot;my/path&quot;) {[], {:file, device}} end @impl Raxx.Server def handle_data(data, state = {:file, device}) do IO.write(device, data) {[], state} end @impl Raxx.Server def handle_tail(_trailers, state) do response(:see_other) |&gt; set_header(&quot;location&quot;, &quot;/&quot;) end end A body may arrive split by packets, chunks or frames. handle_data will be invoked as each part arrives. An application should never assume how a body will be broken into data parts. Request/Response flow It is worth noting what guarantees are given on the request parts passed to the Server&#39;s handle_* functions. It depends on the Server type, Raxx.Server vs Raxx.SimpleServer: So, for example, after a %Raxx.Request{body: false} is passed to a Server&#39;s Raxx.Server.handle_head/2 callback, no further request parts will be passed to to the server (Raxx.Server.handle_info/2 messages might be, though). Similarly, these are the valid sequences of the response parts returned from the Servers: Any Raxx.Middlewares should follow the same logic. Router The Raxx.Router can be used to match requests to specific server modules. defmodule MyApp do use Raxx.Server use Raxx.Router, [ {%{method: :GET, path: []}, HomePage}, {%{method: :GET, path: [&quot;slow&quot;]}, LongPoll}, {%{method: :GET, path: [&quot;messages&quot;]}, SubscribeToMessages}, {%{method: :PUT, path: [&quot;upload&quot;]}, Upload}, {_, NotFoundPage} ] end"}]